module v386
(
    // Тактовый генератор
    input               clock,
    input               reset_n,
    input               locked,
    // Магистраль данных 8 битная
    output      [31:0]  address,
    input       [ 7:0]  in,
    output reg  [ 7:0]  out,
    output reg          we
);

`include "v386_decl.v"

always @(posedge clock)
if (locked) begin // Процессор работает
if (reset_n == 1'b0) begin

    cs          <= 80'h0000_0080_0000_FFFF__F000; // P=1,LIMIT=65535
    eip         <= 16'hFFF0;
    t           <= fetch;
    __adsize    <= 1'b0;
    __opsize    <= 1'b0;
    __override  <= 1'b0;
    __rep       <= 1'b0;
    src         <= 1'b0; // Указатель на CS:EIP
    prot        <= 1'b0; // Процессор работает в REAL MODE

end else case (t)

// Считывание опкода или префикса
fetch: begin

    eip     <= eip_next;
    opcode  <= in;
    size    <= in[0];
    dir     <= in[1];
    alu     <= in[5:3];
    fn      <= 1'b0;
    fn2     <= 1'b0;
    ignoreo <= 1'b0;

    case (in)
    // Сегментные префиксы, 80-разрядные
    8'h26: begin __segment <= es; __override <= 1'b1; end
    8'h2E: begin __segment <= cs; __override <= 1'b1; end
    8'h36: begin __segment <= ss; __override <= 1'b1; end
    8'h3E: begin __segment <= ds; __override <= 1'b1; end
    8'h64: begin __segment <= fs; __override <= 1'b1; end
    8'h65: begin __segment <= gs; __override <= 1'b1; end
    8'h66: begin __opsize <= ~__opsize; end
    8'h67: begin __adsize <= ~__adsize; end
    8'hF0, 8'h9B, 8'h90: begin /* fwait, lock, nop */ end
    8'hF2, 8'hF3: __rep <= in[1:0];
    // Исполнение опкода
    default: begin

        t <= exec;

        // Защелкивание префиксов
        rep         <= __rep;       __rep       <= 2'b00;
        override    <= __override;  __override  <= 1'b0;
        opsize      <= __opsize;    __opsize    <= defsize;
        adsize      <= __adsize;    __adsize    <= defsize;
        segment     <= __segment;   __segment   <= ds;

        // ...

    end
    endcase

end
endcase
end

endmodule
